package creim.analysis.audio.spotifyapi;

import com.sun.net.httpserver.HttpExchange;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.lang.reflect.Field;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;
import javalib.net.http.HTTPExtendedServer;
import javalib.net.http.HTTPUtil;
import javalib.net.http.handler.GetHandler;
import javalib.util.Hash;
import javalib.util.StringUtil;
import creim.Config;
import creim.util.Logger;
import creim.util.Util;
import org.json.simple.JSONArray;
import org.json.simple.JSONObject;
import org.json.simple.parser.JSONParser;
import org.json.simple.parser.ParseException;


/**
 * Spotify API communicator class
 * 
 * @author Viktor
 */
public class SpotifyAPI {
	
	private class URIWrapper {
		public URI codeuri;
	}
	
	public static final File    spotifyCacheRoot = Config.SPOTIFY_ROOT;
	public static boolean       usecache = true;
	
	private final Object        lock;
	private URIWrapper          wrapper;
	private HTTPExtendedServer  callbackserver;
	private String              accessToken;
	private JSONObject          accessJson;
	private JSONParser          jsonparser;
	private File                tokenfile;
	private SpotifyCallbackWeb  spotifyhtml;
	
	
	public SpotifyAPI() {
		lock = new Object();
		wrapper = new URIWrapper();
		jsonparser = new JSONParser();
		tokenfile = new File("spotifytoken.json");
		spotifyhtml = new SpotifyCallbackWeb();
	}
	
	
	public String getAuthCode() throws IOException, URISyntaxException, InterruptedException {
		javalib.JavalibLogger.logger.setLevel(Level.SEVERE);
		callbackserver = new HTTPExtendedServer(8000);
		callbackserver.start();
		callbackserver.addHandler("/callback", new GetHandler() {
			@Override
			public void GET(HttpExchange he) throws IOException {
				synchronized (lock) {
					wrapper.codeuri = he.getRequestURI();
					HTTPExtendedServer.dataResponse(he, spotifyhtml.getBytes(), "text/html");
					lock.notify();
				}
			}
		});
		
		wrapper.codeuri = null;
		String authurl = "https://accounts.spotify.com/authorize?client_id="+Config.SPOTIFY_USR_ID+"&response_type=code&redirect_uri=http://localhost:8000/callback";
		URI authuri = new URI(authurl);
		java.awt.Desktop.getDesktop().browse(authuri);
		
		String code;
		synchronized (lock) {
			lock.wait();
			code = wrapper.codeuri.getQuery();
			callbackserver.stop();
			wrapper.codeuri = null;
		}
		code = code.substring("code=".length());
		
		return code;
	}
	
	
	public JSONObject getStoredTokenJson() {
		try {
			String jsonstring = StringUtil.fromFile(tokenfile);
			JSONObject tokenobj = (JSONObject)jsonparser.parse(jsonstring);
			return tokenobj;
		} catch (IOException | ParseException e) {
			return null;
		}
	}
	
	
	public void storeTokenJson(JSONObject tokenobj) {
		tokenobj.put("created", System.currentTimeMillis());
		try {
			StringUtil.toFile(tokenobj.toJSONString(), tokenfile);
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}
	
	
	public JSONObject getAccessTokenJson() throws IOException, InterruptedException, URISyntaxException, ParseException {
		JSONObject tokenobj;
		
		tokenobj = getStoredTokenJson();
		if (tokenobj != null) {
			long created = (Long)tokenobj.get("created");
			long expires = (Long)tokenobj.get("expires_in");
			long mintime = created + (expires * 1000) - (5 * 60 * 1000);
			long current = System.currentTimeMillis();
			long diff = mintime - current;
			if (diff > 0) {
				Logger.info(String.format("Time until token expires  %d:%d", (int)(diff / 1000 / 60), (int) (diff / 1000 % 60)));
				return tokenobj;
			}
			Logger.info("Token expired");
		}
		
		String url  = "https://accounts.spotify.com/api/token";
		String code = getAuthCode();
		String b64  = Hash.Base64encode(Config.SPOTIFY_CRED.getBytes());
		
		Map<String, String> headers = new HashMap<>();
		headers.put("Authorization", "Basic " + b64);
		
		Map<String, String> data = new HashMap<>();
		data.put("code", code);
		data.put("redirect_uri", "http://localhost:8000/callback");
		data.put("grant_type", "authorization_code");
		
		StringBuilder parameters = new StringBuilder();
		for (String key : data.keySet()) {
			parameters.append(key);
			parameters.append("=");
			parameters.append(data.get(key));
			parameters.append("&");
		}
		parameters.replace(parameters.length()-1, parameters.length(), "");
		
		HTTPUtil.POSTResponse res = HTTPUtil.post(url, parameters.toString(), headers);
		
		tokenobj = (JSONObject)jsonparser.parse(res.response);
		storeTokenJson(tokenobj);
		
		return tokenobj;
	}
	
	
	public JSONObject apiCallAuth(String url) throws IOException {
		return apiCallAuth(url, true);
	}
	
	public JSONObject apiCallAuth(String url, boolean cache) throws IOException {
		String data = null;
		data = getCached(url);

		if (data != null) {
			try {
				return (JSONObject)jsonparser.parse(data);
			} catch (Exception e) {}
		}
		Logger.notify("API Call: " + url);
		Map<String, String> headers = new HashMap<>();
		headers.put("Authorization", "Bearer "+getAccessToken());
		
		String text = HTTPUtil.getText(url, headers);
		JSONObject jsonobj = null;
		try {
			jsonobj = (JSONObject)jsonparser.parse(text);
		} catch (ParseException e) {}
		
		if (cache) {
			cache(url, Util.json(jsonobj));
		}
		
		return jsonobj;
	}
	
	
	private void cache(String url, String data) {
		String[] parts = url.split("/");
		String trackid = parts[parts.length - 1];
		String endpoint = parts[parts.length - 2];
		
		File trackdir = new File(spotifyCacheRoot, trackid);
		if (!trackdir.exists())
			trackdir.mkdirs();
		
		String filename = endpoint + ".json";
		File cachefile = new File(trackdir, filename);
		if (!cachefile.exists()) {
			try {
				StringUtil.toFile(data, cachefile);
			} catch (Exception e) {
				Logger.warning("Failed to cache: "+cachefile);
			}
		}
	}
	
	
	public String getCached(String url) {
		String[] parts = url.split("/");
		String trackid = parts[parts.length - 1];
		String endpoint = parts[parts.length - 2];
		File trackdir = new File(spotifyCacheRoot, trackid);
		
		String filename = endpoint + ".json";
		File cachefile = new File(trackdir, filename);
		if (cachefile.exists()) {
			try {
				String data = StringUtil.fromFile(cachefile);
				Logger.debug("Cache file:"+cachefile+" OK.");
				return data;
			} catch (Exception e) {}
		}
		Logger.debug("Cache file:"+cachefile+" NOT FOUND.");
		return null;
	}
	
	
	public String getAccessToken() {
		if (accessToken == null) {
			if (accessJson == null) {
				try {
					accessJson = getAccessTokenJson();
					accessToken = (String)accessJson.get("access_token");
				} catch (IOException | InterruptedException | URISyntaxException | ParseException e) {
					throw new RuntimeException("Spotify API ERROR: "+e.getMessage());
				}
			} else {
				accessToken = (String)accessJson.get("access_token");
			}
		}
		return accessToken;
	}
	
	
	public void setAccessToken(String token) {
		this.accessToken = token;
	}
	
	
	
	public JSONObject getTrackById(String trackid) throws IOException {
		trackid = getTrackID(trackid);
		String url = "https://api.spotify.com/v1/tracks/"+trackid;
		return apiCallAuth(url);
	}
	
	
	public String getTrackName(String trackid) throws IOException {
		JSONObject trackjson = getTrackById(trackid);
		return trackjson.get("name").toString();
	}
	
	
	public String getTrackArtist(String trackid) throws IOException {
		JSONObject trackjson = getTrackById(trackid);
		return ((JSONObject)((JSONArray)trackjson.get("artists")).get(0)).get("name").toString();
	}
	
	
	public JSONObject getAlbumByTrackId(String trackid) throws IOException {
		trackid = getTrackID(trackid);
		JSONObject track = getTrackById(trackid);
		String albumid = (String)(((JSONObject)track.get("album")).get("id"));
		String url = "https://api.spotify.com/v1/albums/"+albumid;
		return apiCallAuth(url, false);
	}
	
	
	public JSONObject getAlbumById(String albumid) throws IOException {
		String url = "https://api.spotify.com/v1/albums/"+albumid;
		return apiCallAuth(url);
	}
	
	
	public JSONObject getAudioFeatuers(String trackid) throws IOException {
		trackid = getTrackID(trackid);
		String url = "https://api.spotify.com/v1/audio-features/"+trackid;
		return apiCallAuth(url);
	}
	
	
	public JSONObject getAnalysis(String trackid) throws IOException {
		trackid = getTrackID(trackid);
		String url = "https://api.spotify.com/v1/audio-analysis/"+trackid;
		return apiCallAuth(url);
	}
	
	
	public SpotifyAudioFeatures getAudioFeatureObject(String trackid) throws IOException {
		trackid = getTrackID(trackid);
		
		JSONObject features = getAudioFeatuers(trackid);
		SpotifyAudioFeatures af = new SpotifyAudioFeatures();
		for (Field field : af.getClass().getFields()) {
			try {
				field.set(af, features.get(field.getName()));
			} catch (Exception e) {}
		}
		
//		JSONObject albums = getAlbumByTrackId(trackid);
//		JSONArray genres = (JSONArray)albums.get("genres");
//		if (genres != null) {
//			af.genres = (String[])genres.toArray(new String[genres.size()]);
//		}
		return af;
	}
	
	
	public boolean hasAllDataStored(String trackid) {
		return hasAllDataStored(trackid, spotifyCacheRoot);
	}
	
	public boolean hasAllDataStored(String trackid, File outputdir) {
		trackid = getTrackID(trackid);
		File trackdir = new File(outputdir, trackid);
		if (!trackdir.exists()) {
			return false;
		}
		
		File track = new File(trackdir, "tracks.json");
		if (!track.exists()) {
			return false;
		}
		
		File features = new File(trackdir, "audio-features.json");
		if (!features.exists()) {
			return false;
		}
		
		File analysis = new File(trackdir, "audio-analysis.json");
		if (!analysis.exists()) {
			return false;
		}
		
		return true;
	}
	
	public boolean hasFile(String trackid, String filename) {
		trackid = getTrackID(trackid);
		File trackdir = new File(spotifyCacheRoot, trackid);
		if (!trackdir.exists()) {
			return false;
		}
		File file = new File(trackdir, filename);
		return file.exists();
	}
	
	
	public File fetchAllAndStore(String trackid) throws IOException {
		return fetchAllAndStore(trackid, spotifyCacheRoot);
	}
	
	public File fetchAllAndStore(String trackid, File outputdir) throws IOException {
		trackid = getTrackID(trackid);
		File trackdir = new File(outputdir, trackid);

		Logger.info("Getting tracks.json for "+trackid);
		JSONObject track = getTrackById(trackid);
		
		if (!trackdir.exists()) {
			trackdir.mkdirs();
		}
		StringUtil.toFile(Util.json(track), new File(trackdir, "tracks.json"));
		
		
		Logger.info("Getting audio-features.json for "+trackid);
		JSONObject features = getAudioFeatuers(trackid);
		StringUtil.toFile(Util.json(features), new File(trackdir, "audio-features.json"));
		
		
		Logger.info("Getting audio-analysis.json for "+trackid);
		JSONObject analysis = getAnalysis(trackid);
		StringUtil.toFile(Util.json(analysis), new File(trackdir, "audio-analysis.json"));
		
		
//		Logger.info("Getting albums.json ... ");
//		JSONObject album = null;
//		try { 
//			album = getAlbumByTrackId(trackid); 
//			StringUtil.toFile(Util.json(album), new File(trackdir, "albums.json"));
//		} catch (IOException e) {
//			Logger.notify("Missing album info ... ");
//		}
		
		
		String previewurl = (String)track.get("preview_url");
		File file = new File(trackdir, trackid + ".mp3");
		if (!file.exists()) {
			Logger.info("Getting preview song for "+trackid);
			byte[] mp3data = HTTPUtil.get(previewurl);
			try (FileOutputStream fos = new FileOutputStream(file)) {
				fos.write(mp3data);
			}
		}
		
		return trackdir;
	}
	
	
	public String getTrackID(String trackid) {
		if (trackid.startsWith("https://open.spotify.com/track/")) {
			trackid = trackid.substring("https://open.spotify.com/track/".length());
			return trackid;
		}
		
		int i;
		if ((i = trackid.lastIndexOf(':')) != -1) {
			trackid = trackid.substring(i+1);
		}
		return trackid;
	}
}
