package creim.analysis.image;

import java.awt.Color;
import java.awt.image.BufferedImage;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;
import org.apache.commons.math3.ml.clustering.CentroidCluster;
import org.apache.commons.math3.ml.clustering.Clusterable;
import org.apache.commons.math3.ml.clustering.KMeansPlusPlusClusterer;
import org.apache.commons.math3.ml.distance.EuclideanDistance;

public class ImageAnalyzer {

	private BufferedImage img;

	public ImageAnalyzer(BufferedImage img) {
		this.img = img;
	}

	
	/**
	 * Returns arithmetic mean color
	 * @return 
	 */
	public Color meanColor() {
		long r = 0;
		long g = 0;
		long b = 0;
		long a = 0;
		long pc = img.getWidth() * img.getHeight();

		for (int y = 0; y < img.getHeight(); y++) {
			for (int x = 0; x < img.getWidth(); x++) {
				int c = img.getRGB(x, y);
				a += (c >>> 24) & 0xff;
				r += (c >>> 16) & 0xff;
				g += (c >>>  8) & 0xff;
				b += (c >>>  0) & 0xff;
			}
		}
		int ai = (int)(a / pc);
		int ri = (int)(r / pc);
		int gi = (int)(g / pc);
		int bi = (int)(b / pc);
		return new Color(ri, gi, bi, 255);
	}
	
	
	
	public double colorVariation() {
		Color mean = meanColor();
		float[] hsbvals = new float[3];
		Color.RGBtoHSB(mean.getRed(), mean.getGreen(), mean.getBlue(), hsbvals);
		
		System.out.println(Arrays.toString(hsbvals));
		
		return 0;
	}
	
	
	/**
	 * Returns color palette using k-means clustering
	 * @return 
	 */
	public Color[] palette() {
		int n = 5;
		List<ColorPoint> colorpoints = new LinkedList<>();
		for (int y = 0; y < img.getHeight(); y++) {
			for (int x = 0; x < img.getWidth(); x++) {
				colorpoints.add(new ColorPoint(new Color(img.getRGB(x, y))));
			}
		}
		
		KMeansPlusPlusClusterer<ColorPoint> clusterer = new KMeansPlusPlusClusterer<>(n, 10000, new EuclideanDistance());
		List<CentroidCluster<ColorPoint>> clusterResults = clusterer.cluster(colorpoints);
		
		Color[] palette = new Color[n];
		for (int i = 0; i < n; i++) {
			long r = 0;
			long g = 0;
			long b = 0;
			int  t = 0;
			for (ColorPoint cw : clusterResults.get(i).getPoints()) {
				double[] p = cw.getPoint();
				r += (int)p[0];
				g += (int)p[1];
				b += (int)p[2];
				t++;
			}
			palette[i] = new Color((int)(r / t), (int)(g / t), (int)(b / t));
		}
		return palette;
	}
	private static class ColorPoint implements Clusterable {
		private double[] colors;
		public ColorPoint(Color c) {
			this.colors = new double[]{ c.getRed(), c.getGreen(), c.getBlue() };
		}
		@Override
		public double[] getPoint() {
			return colors;
		}
	}
	
	
	/**
	 * 
	 * @param color
	 * @param D acceptable distance from input color of each color component
	 * @return 
	 */
	public double percentOfColor(Color color, int D) {
		return percentOfColor(color, D, D, D, D);
	}
	
	
	/**
	 * Calculates percentage of input color in image.
	 * Integer parameters defines the acceptable distance of each color component in image, from input color.
	 * i.e if dr is 100 the red component can be within +-100 values to approve presence of color
	 * @param color
	 * @param da alpha range
	 * @param dr red range
	 * @param dg green range
	 * @param db blue range
	 * @return percentage 0-1
	 */
	public double percentOfColor(Color color, int da, int dr, int dg, int db) {
		int ia = color.getAlpha();
		int ir = color.getRed();
		int ig = color.getGreen();
		int ib = color.getBlue();
		final double tot = img.getHeight() * img.getWidth();
		int count = 0;
		
		for (int y = 0; y < img.getHeight(); y++) {
			for (int x = 0; x < img.getWidth(); x++) {
				int c = img.getRGB(x, y);
				int a = (c >>> 24) & 0xff;
				int r = (c >>> 16) & 0xff;
				int g = (c >>>  8) & 0xff;
				int b = (c)        & 0xff;
				
				boolean ok = ia-da <= a && a <= ia+da &&
				             ir-dr <= r && r <= ir+dr &&
				             ig-dg <= g && g <= ig+dg &&
				             ib-db <= b && b <= ib+db ;
				
				count += ok ? 1 : 0;
			}
		}
		
		return count / tot;
	}
	
	
	
}
