package creim.art.painter;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.image.BufferedImage;
import javalib.util.MathUtil;
import javalib.util.RandomUtil;
import creim.graphics.colors.ColorUtil;
import creim.art.painter.Painter;
import java.awt.Dimension;

public class CircleBrushPainter extends Painter {
	
	/** Array of brush point coordinates used to simulate brush hairs */
	private int[][] brushpoints;
	/** Amount of brush hairs */
	private int brushhairs;
	/** X coordinate */
	public int x;
	/** Y coordinate */
	public int y;
	/** Width of the oval */
	public int w;
	/** Height of the oval */
	public int h;
	/** Width of the brush */
	public float brushwidth;
	/** Color of the brush */
	public Color brushcolor;

	public CircleBrushPainter() {
	}
	
	/**
	 * Paints ovals/circles using simulated brush strokes.
	 * @param x center x coordinate
	 * @param y center y coordinate
	 * @param w oval width
	 * @param h oval height
	 * @param brushwidth brush width used to draw
	 * @param brushcolor RGB brush color
	 */
	public CircleBrushPainter(int x, int y, int w, int h, float brushwidth, Color brushcolor) {
		this.x = x;
		this.y = y;
		this.w = w;
		this.h = h;
		this.brushwidth = brushwidth;
		this.brushcolor = brushcolor;
		
		this.brushhairs = (int)(this.brushwidth * Math.log(this.brushwidth) * 10);
		this.brushpoints = new int[brushhairs][2];

		for (int i = 0; i < brushhairs; i++) {
			double t = 2 * Math.PI * RandomUtil.nextDouble();
			double u = RandomUtil.nextDouble() + RandomUtil.nextDouble();
			double r = u > 1 ? 2-u : u;
			int px = (int)((Math.cos(t) * r) * this.brushwidth);
			int py = (int)((Math.sin(t) * r) * this.brushwidth);

			int[] p = {px, py};
			brushpoints[i] = p;
		}
	}

	@Override
	public void randomize() {
		x = RandomUtil.nextInt(-margin, imagedim.width + margin);
		y = RandomUtil.nextInt(-margin, imagedim.height + margin);
		w = RandomUtil.nextInt(-margin, imagedim.width + margin);
		h = RandomUtil.nextInt(-margin, imagedim.height + margin);
		brushcolor = palette[RandomUtil.nextInt(palette.length)];
		brushwidth = RandomUtil.nextInt(75);
	}
	
	
	
	@Override
	public void mutate(Dimension imagedim, Color[] palette, int margin) {
		int p = RandomUtil.nextInt(6);
		if (p == 0)
			x = RandomUtil.nextInt(-margin, imagedim.width+margin);
		else if (p == 1)
			y = RandomUtil.nextInt(-margin, imagedim.height+margin);
		else if (p == 2)
			w = RandomUtil.nextInt(-margin, imagedim.width+margin);
		else if (p == 3)
			h = RandomUtil.nextInt(-margin, imagedim.height+margin);
		else if (p == 4)
			brushcolor = palette[RandomUtil.nextInt(palette.length)];
		else if (p == 5)
			brushwidth = RandomUtil.nextInt(75);
	}




	@Override
	public BufferedImage paint() {
		throw new RuntimeException("Broken");
	}

	@Override
	public void paintOn(BufferedImage image) {
		Graphics2D g2 = (Graphics2D) image.getGraphics();
		g2.setColor(brushcolor);
		g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
		
		double[] brushstrength = new double[brushhairs];
		double[] brusheffect = new double[brushhairs];
		for (int i = 0; i < brushhairs; i++) {
			int[] p = brushpoints[i];
			double dist = Math.sqrt(p[0] * p[0] + p[1] * p[1]);
			brushstrength[i] = dist > 0.01 ? 1.0 / dist : 0.01;
			brusheffect[i] = RandomUtil.RNG.nextDouble() * 0.2;// * dist;
		}
		

		int steps = 500;
		
		for (double s = 0; s < steps-1; s++) {
			double x0 = Math.cos(s/steps*2*Math.PI) * w + x + w / 2.0;
			double y0 = Math.sin(s/steps*2*Math.PI) * h + y + h / 2.0;
			double x1 = Math.cos((s+1)/steps*2*Math.PI) * w + x + w / 2.0;
			double y1 = Math.sin((s+1)/steps*2*Math.PI) * h + y + h / 2.0;

			for (int i = 0; i < brushhairs; i++) {
				double ef = ((Math.sin(brusheffect[i] * s) + 1.0) / 2.0) * brushstrength[i];
				int rgb = (brushcolor.getRGB() & 0x00ffffff);
				rgb = ColorUtil.scale(rgb, 1.0 + (ef + RandomUtil.nextGaussian(0.0, 0.05)));

				Color c = new Color(( (rgb & 0x00ffffff) | ((int)(MathUtil.scale(Math.min(ef, 1.0), 0, 1) * 255) << 24)), true);

				g2.setColor(c);
				
				int[] p = brushpoints[i];
				g2.drawLine((int)(p[0] + x0), (int)(p[1] + y0), (int)(p[0] + x1), (int)(p[1] + y1));
			}
			
			for (int i = 0; i < brushhairs; i++) {
				brushstrength[i] *= 0.999;
			}
		}

		g2.dispose();
	}

	@Override
	public double[] centerPoint() {
		return new double[]{x + w/2.0, y + h/2.0};
	}

	@Override
	public double area() {
		return w/2.0 * h/2.0 * Math.PI;
	}
}

