package creim.graphics;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.image.BufferedImage;
import javalib.util.ColorUtil;
import creim.analysis.audio.AudioFileAnalyzer;
import creim.util.Logger;
import org.apache.commons.math3.analysis.interpolation.PiecewiseBicubicSplineInterpolatingFunction;
import org.apache.commons.math3.analysis.interpolation.PiecewiseBicubicSplineInterpolator;
import org.apache.commons.math3.stat.StatUtils;



public class SpectrumImageLight {
	
	
	public static final Color BACKGROUND_COLOR = Color.WHITE;
	public static final Color FOREGROUND_COLOR = Color.BLACK;
	public static final Color AVG_LINE_COLOR = Color.BLACK;
	public static final Color AVG_AMP_PER_TIME_COLOR = Color.RED.darker();
	public static final Color AVG_AMP_PER_FREQ_COLOR = Color.CYAN.darker();
	public static final float[][] HEATMAP_COLORS = new float[][]{ {0,0,0,0.2f}, {0,0,0.75f,1}, {0,0.75f,0,1f}, {0.75f,0.75f,0,1}, {0.75f,0,0,1}, {0.5f,0,0,1} };
	
	
	public static void createInterpolated(BufferedImage img, Graphics2D g2, int sx, int sy, int w, int h, double[][] data) {
		
		// Max value ====================================================================================================
		double maxvalue = 0;
		for (int i = 0; i < data.length; i++) {
			for (int j = 0; j < data[i].length; j++) {
				if (data[i][j] > maxvalue) {
					maxvalue = data[i][j];
				}
			}
		}
		
		// Interpolation ================================================================================================
		double[] xval = new double[data.length];
		for (int i = 0; i < xval.length; i++) {
			xval[i] = i;
		}
		double[] yval = new double[data[0].length];
		for (int i = 0; i < yval.length; i++) {
			yval[i] = i;
		}
		
		// Spectrum image ===============================================================================================
		PiecewiseBicubicSplineInterpolator biinterpolator = new PiecewiseBicubicSplineInterpolator();
		PiecewiseBicubicSplineInterpolatingFunction function;
		function = biinterpolator.interpolate(xval, yval, data);
		for (int x = 0; x < w; x++) {
			for (int y = 0; y < h; y++) {
				double fx = (x / (double) w) * (data.length-1);
				double fy = (y / (double) h) * (data[0].length-1);
				
				Color c = ColorUtil.heatmap(function.value(fx, fy), 0, maxvalue, HEATMAP_COLORS);
				g2.setColor(c);
				g2.fillRect(sx + x, h - (sy + y), 1, 1);
//				img.setRGB(sx + x, h - (sy + y), c.getRGB());
			}
		}
	}
	
	
	
	public static void createCubed(BufferedImage img, Graphics2D g2, int sx, int sy, int w, int h, double[][] data) {
		// Max value ====================================================================================================
		double maxvalue = 0;
		for (int i = 0; i < data.length; i++) {
			for (int j = 0; j < data[i].length; j++) {
				if (data[i][j] > maxvalue) {
					maxvalue = data[i][j];
				}
			}
		}
		
		double cw = w / (double)data.length;
		double ch = h / (double)data[0].length;
//		double cw = 2;
//		double ch = 2;
		
		for (int x = 0; x < data.length; x++) {
			for (int y = 0; y < data[x].length; y++) {
				double cx = x * cw;
				double cy = y * ch;
				
				Color c = ColorUtil.heatmap(data[x][y], 0, maxvalue);
				c = c.darker().darker().darker();
				//c = new Color(c.getRed(), c.getGreen(), c.getBlue(), 100);
				g2.setColor(c);
				g2.fillRect((int)Math.ceil(cx), h - (int)Math.ceil(cy) - (int)Math.ceil(ch), (int)Math.ceil(cw), (int)Math.ceil(ch));
			}
		}
	}



	public static BufferedImage create(int w, int h, double[][] data) {
		return create(w, h, data, true);
	}
	
	public static BufferedImage create(int w, int h, double[][] data, boolean interpolated) {
		Logger.info("Creating spectrum image ... ");
		int ho = h;
		int wo = w;
		int fontsize = 10;
		int offset = 150;

		w -= offset;
		h -= offset;
		AudioFileAnalyzer an = new AudioFileAnalyzer();
		
		BufferedImage img = new BufferedImage(wo, ho, BufferedImage.TYPE_INT_ARGB);
		Graphics2D g2 = img.createGraphics();
		g2.setColor(BACKGROUND_COLOR);
		g2.fillRect(0, 0, wo, ho);
		
		if (interpolated)
			createInterpolated(img, g2, 0, 0, w, h, data);
		else
			createCubed(img, g2, 0, 0, w, h, data);
		
		
		// Data points text =============================================================================================
		g2.setFont(new Font("Arial", Font.PLAIN, fontsize));
		g2.setColor(FOREGROUND_COLOR);
		for (int x = 0; x < w; x += 50) {
			double fx = (x / (double) w)  * (data.length-1);
			g2.drawString(String.format("%.0f",fx), x, ho);
		}
		
		
		// AVG Frequencies ==============================================================================================
		g2.setColor(AVG_LINE_COLOR);
		g2.setStroke(new BasicStroke(2.0f));
		g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
		
		double[] avg = an.meanFrequencies(data);
		double sx = 1;
		double sy = avg[1];
		int sfx = (int)((sx / (double) (avg.length)) * w);
		int sfy = (int)((sy / (double) (data[0].length)) * h);
		int lx = sfx;
		int ly = sfy;
		
		for (int x = 2; x < avg.length; x++) {
			int fx = (int)((x / (double) (avg.length)) * w);
			
			double y = avg[x];
			int fy = (int)((y / (double) (data[0].length)) * h);
			
			g2.drawLine(lx, h - ly, fx, h - fy);
			lx = fx;
			ly = fy;
		}
		g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_OFF);
		g2.setStroke(new BasicStroke(1.0f));
		
		
		// AVG Amp across time ==========================================================================================
		g2.setColor(AVG_AMP_PER_TIME_COLOR);
		double[] avgamptime = an.meanAmplitudesPerTime(data);
		double avgamptimemax = StatUtils.max(avgamptime);
		double avgamptimemin = StatUtils.min(avgamptime);
		double avgamptimescale = (ho-h-fontsize) / (avgamptimemax - avgamptimemin);
		
		for (int x = 0; x < avgamptime.length; x++) {
			double amp = avgamptime[x];
			double ampheight = (int)(amp * avgamptimescale) - avgamptimemin;
			double ampwidth = w / (double)avgamptime.length;
			double startx = ampwidth * x;
			
			g2.fillRect((int)startx, (int)(ho-ampheight-fontsize), (int)Math.ceil(ampwidth), (int)(ampheight));
		}
		
		
		{
			double avgavgamptime = StatUtils.mean(avgamptime);
			double ampheight = (int)(avgavgamptime * avgamptimescale) - avgamptimemin;
			double ampwidth = w / (double)avgamptime.length;
			g2.setStroke(new BasicStroke(2.0f));
			g2.setColor(AVG_LINE_COLOR);
			g2.drawLine(0, (int)(ho-ampheight-fontsize), w, (int)(ho-ampheight-fontsize));
		}
		
		
		
		// AVG Amp across freq ==========================================================================================
		g2.setColor(AVG_AMP_PER_FREQ_COLOR);
		double[] avgampfreq = an.meanAmplitudesPerFrequency(data);
		for (int y = 0; y < avgampfreq.length; y++) {double amp = avgampfreq[y];
			double ampheight = h / (double)avgampfreq.length;
			double ampwidth = (int)(amp * 3.5);
			double starty = ampheight * y;
			
			g2.fillRect((int)(wo-ampwidth), h - (int)starty - (int)Math.ceil(ampheight),  (int)ampwidth, (int)Math.ceil(ampheight));
		}
		

		
		return img;
	}
	
}
